//! Standard extensions to [`crate::core`].

pub mod atomic;
pub mod cast;
pub mod collections;
pub mod fallible;
pub mod inlining;
mod local_origin;
pub mod nullable;
pub mod point;
pub mod tracing;
mod typeless;

use std::{error::Error, fmt::Display, rc::Rc};

use crate::core::*;
use crate::func::*;

impl Display for Address {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}@{}", hex::encode(self.point), self.index))
    }
}

/// Extension trait for [Serializable]s.
pub trait ExtSerializable: Serializable {
    /// Serialize into a [Vec] of bytes.
    fn bytes(&self) -> Vec<u8>;
}

impl<S: Serializable> ExtSerializable for S {
    fn bytes(&self) -> Vec<u8> {
        let mut vec = Vec::new();
        self.serialize(&mut vec);
        vec
    }
}
